import atexit
from pathlib import Path
from typing import Dict, Union

from .template import MemoryTemplate

DICT_EXPOSED_METHODS = ['clear', 'items']


class MemoryEngine(object):
    _data: Dict
    _location: Path
    _template: MemoryTemplate

    def __init__(self, location: Union[Path, str], template: MemoryTemplate, auto_load=True):
        """
        :param location: path to save file
        :param template: memory template
        """

        self._data = {}
        self._template = template

        # update location
        self.location = location

        # exposing dictionary methods
        self.clear = self._data.clear
        self.items = self._data.items

        # for method in DICT_EXPOSED_METHODS:
        #     setattr(self, method, getattr(self._data, method))

        # read the initial data

        if auto_load:
            self.load()

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value: Union[Path, str]):
        # conversion
        if type(value) != Path:
            self._location = Path(value)
        else:
            self._location = value

        # making sure directory exists
        self._location.parent.mkdir(parents=True, exist_ok=True)

        # updating template location attribute
        self._template.location = self._location

    @property
    def template(self):
        return self._template

    def save(self):
        """
        write current data to disk
        """
        self._template.save(self._data)

    def load(self):
        """
        read data from disk
        """
        try:
            self._data = self._template.load()
        except FileNotFoundError:
            self.save()

    def get(self, key, default=None):
        """
        :param key: key used as identifier
        :param default: value to return is key not found

        :return: data corresponding to identifer(key)
        :returns: default if key not found
        """
        try:
            value = self._data[key]
        except KeyError:
            value = default

        return value

    def delete(self, *args):
        """
        removes the keys from memory

        :param args: keys to be removed
        """
        for key in args:
            try:
                del self._data[key]
            except KeyError:
                pass

    def put(self, key, value):
        """
        adds key-value pair to memory

        :param key: key used as identifier
        :param value: data to store
        :return: self, may be chained
        """
        self._data[key] = value

        return self

    def putall(self, d: dict):
        """
        adds all the key-value pairs in the map

        :param d: dictionary map to be stored
        """
        for key, value in d.items():
            self._data[key] = value

    def save_atexit(self, should_save=True):
        """
        register save function to atexit module

        :param should_save: whether to register or unregister
        """
        if should_save:
            atexit.register(self.save)
        else:
            atexit.unregister(self.save)


import json

from ..template import MemoryTemplate


class JsonTemplate(MemoryTemplate):
    """
    Memory template using json storage
    """

    def save(self, data: dict):
        with self.location.open('w') as f:
            json.dump(data, f)

    def load(self) -> dict:
        with self.location.open('r') as f:
            return json.load(f)



import pickle
import unittest
from pathlib import Path

from xmem import MemoryEngine
from xmem.templates import PickleTemplate

TEST_PATH = 'pickletest'
DUMMY_DATA = {f'test {i}': i for i in range(10)}


class JsonTemplateTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.memory = MemoryEngine(TEST_PATH, PickleTemplate())

    def tearDown(self) -> None:
        self.memory._data.clear()
        self.memory._location.unlink()

    def test_save(self):
        self.memory._data = DUMMY_DATA

        self.memory.save()

        with Path(TEST_PATH).open('rb') as f:
            self.assertEqual(pickle.load(f), DUMMY_DATA)

    def test_load(self):
        with Path(TEST_PATH).open('wb') as f:
            pickle.dump(DUMMY_DATA, f)

        self.memory.load()

        self.assertEqual(self.memory._data, DUMMY_DATA)


import unittest
from pathlib import Path

from xmem import MemoryEngine, MemoryTemplate


class MemoryEngineTestCase(unittest.TestCase):
    def setUp(self) -> None:
        # template left as EmptyTemplate, as data wont be written to disk
        # auto_load disabled since template left empty
        self.memory = MemoryEngine('testdata', MemoryTemplate(), auto_load=False)

    def tearDown(self) -> None:
        self.memory._data.clear()

    def test_string_location_to_path(self):
        LOCATION = 'memorychanged'
        PATH = Path(LOCATION)

        self.memory.location = LOCATION

        self.assertEqual(self.memory._location, PATH)
        self.assertEqual(self.memory._template.location, PATH)

    def test_put(self):
        KEY = 'put'
        VALUE = 'PVALUE'

        self.memory.put(KEY, VALUE)

        self.assertEqual(self.memory._data[KEY], VALUE)

    def test_putall(self):
        DATA = {str(i): i for i in range(5)}

        self.memory.putall(DATA)

        for key, value in DATA.items():
            self.assertEqual(self.memory._data[key], value)

    def test_get(self):
        KEY = 'get'
        VALUE = 'GVALUE'

        self.memory._data[KEY] = VALUE

        self.assertEqual(self.memory.get(KEY), VALUE)

    def test_delete(self):
        KEY = 'delete'
        VALUE = 'DVALUE'

        self.memory._data[KEY] = VALUE

        self.memory.delete(KEY)

        with self.assertRaises(KeyError):
            _ = self.memory._data[KEY]


import pickle

from ..template import MemoryTemplate


class PickleTemplate(MemoryTemplate):
    """
    Memory template using pickle storage
    """

    def save(self, data: dict):
        with self.location.open('wb') as f:
            pickle.dump(data, f)

    def load(self) -> dict:
        with self.location.open('rb') as f:
            return pickle.load(f)



from setuptools import setup, find_packages
import xmem

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='xmem',
    version=xmem.__version__,
    author='Schicksal',
    description='A simple, light, easy-to-use, and extensible memory module',
    author_email='mhaisham79@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
    ],

    license="MIT license",
    keywords='memory storage json pickle extensible light',

    url='https://github.com/mHaisham/xmem',
    packages=find_packages(),
    python_requires='>=3.6'
)



from pathlib import Path


class MemoryTemplate:
    location: Path

    def save(self, data: dict):
        """
        write the given dictionary to :location:

        :param data: item to save
        """
        raise NotImplementedError

    def load(self) -> dict:
        """
        :return: content read from disk
        """
        raise NotImplementedError
